# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  numpy                                                       as  np
from    sys      import  path                                       as  sys_path
from    os.path  import  abspath, dirname, basename, isfile, isdir
from    os.path  import  join                                       as  pjoin
from    os       import  mkdir                                      as  os_mkdir
from    os       import  listdir                                    as  os_listdir
from    os       import  system                                     as  os_system
from    os       import  walk                                       as  os_walk
from    os.path  import  relpath                                    as  prelpath
from    sys      import  argv
from    copy     import  deepcopy
import  zipfile

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

sys_path.append(dirname(abspath(__file__)))
from  dns_variables  import lmap, fill_geom_vars, make_rough_surface, make_rough_surface_ghost, global_stack
sys_path.pop()

# ------------------------------------------------------------------------
#                  Basic Functions
# ------------------------------------------------------------------------

lfilter               =  lambda f,x: list(filter(f,x))
listdir_full          =  lambda x: sorted(pjoin(x,y) for y in os_listdir(x))
listdir_full_files    =  lambda x: lfilter(isfile, listdir_full(x))
listdir_full_folders  =  lambda x: lfilter(isdir , listdir_full(x))

def apply_include(fname):
    A     = reader(fname)
    all_i = [None]
    while all_i:
        
        all_i = [i for i,x in enumerate(A) if x.lstrip(' ')[:7] == 'include']

        for i in all_i[::-1]:
            _,new,__  =  A[i].replace('"',"'").split("'")
            pad       =  ' '*(len(A[i]) - len(A[i].lstrip(' ')))
            A[i]      =  pad + '! ' + A[i].lstrip(' ')
            A         = A[:i+1] + reader(pjoin(dirname(fname),new)) + A[i+1:]
    writer(fname, A)

def get_tag(A, tag, begin='=', end=')'):
    x = locate_one(tag, A).split(tag)[1]
    if begin:
        x = x.split(begin)[1]
    if end:
        x = x.split(end)[0]
    return x

locate_one  =  lambda tag, A: pop1(lfilter(lambda x: tag in x, A))
locate_ind  =  lambda tag, A: pop1([i for i,x in enumerate(A) if (tag in x)])

def reader(fname):
    with open(fname,'r') as f:
        return [x.rstrip('\n') for x in f]

deepdirname  =  lambda x,n: x if n<1 else deepdirname(dirname(x),n-1)

def apply_no_comment(fname):
    A = reader(fname)
    for i,x in enumerate(A):
        A[i] = x.split('! ')[0]
    A = list(filter(lambda x: bool(x.lstrip(' ')), A))
    A = ['! ---------------------- CODE AUTOMATICALLY GENERATED BY PYTHON ----------------------'] + A
    writer(fname, A)

def pop1(A):
    x, = A 
    return x

def mkdir_p(x):
    assert not isfile(x)
    if not isdir(x):
        mkdir_p(dirname(x))
        os_mkdir(x)
    assert isdir(x)

def import_pyjson(fname, comment = '#'):
    A = reader(fname)
    for i in range(len(A)):
        A[i] = A[i].split(comment)[0]
    result = eval(' '.join(A))
    assert not any((type(x)==str) for x in result.values())
    return result

def resolve_define(A, defined,must_find=True):
    f_inds           =  lambda tag: np.array([i for i,x in enumerate(A) if tag in x]).astype(int)
    i_def            =  f_inds('#IF')
    i_else           =  f_inds('#ELSE' )
    i_end            =  f_inds('#ENDIF')

    levels           = np.zeros(len(A)+1).astype(int) # added extra position to avoid out-of-bounds (doesn't matter)
    levels[i_def  ] +=  1
    levels[i_end+1] += -1
    levels           = np.cumsum(levels)

    assert len(i_def) == len(i_end)
    comment = lambda x: ' '*(len(x)-len(x.lstrip(' '))) + '! ' + x.lstrip(' ')

    assert all( (sum((key in x) for x in defined.keys())==1) for key in defined.keys() )
    
    for macro,value in defined.items():
        i_start = [i for i in i_def if macro in A[i]]
        
        if must_find:
            assert i_start
        
        for i in reversed(i_start):
            j          =  lfilter(lambda j: i     < j  and  levels[i]==levels[j], i_end )[0]
            k          =  lfilter(lambda k: i < k < j  and  levels[i]==levels[k], i_else)

            found_else = bool(len(k))
            if found_else:
                k, = k
            
            to_comment = [i,j]
            if found_else:
                to_comment.append(k)
            
            if value:
                # comment else-range (if any)
                if found_else:
                    to_comment.extend(list(range(k,j+1)))
            else:
                # comment if-range (up to else or end)
                last = k if found_else else j
                to_comment.extend(list(range(i,last+1)))
            for k in sorted(set(to_comment)):
                A[k] = comment(A[k])
            
    # assert not any(('#' in x) for x in A)
    return A

def quick_replacer(fname, holders, defined):
    A = reader(fname)
    
    for i in reversed(range(len(A))):
        for key,val in sorted(holders.items(), key=lambda kv: [-len(kv[0]), kv]):
            assert 'PYTHON_HOLDER_' == key[:14]
            if key in A[i]:
                if (type(val) in [list,tuple]) or hasattr(val,'shape'):
                    if hasattr(val,'shape'):
                        val = val.tolist()
                    pad = ' '*(len(A[i])-len(A[i].lstrip(' ')))
                    A = A[:i] + lmap(lambda c: pad+(str(c)), val) + A[i+1:]
                else:
                    A[i] = A[i].replace(key,str(val))
    if defined:
        A = resolve_define(A, defined)
    writer(fname, A)

def writer(fname, A):
    with open(fname,'w') as f:
        for x in A:
            f.write(x.rstrip('\n') + '\n')

def get_coords_pu(L, n):
    assert type(L)==float and type(n)==int
    xu_full  =  np.linspace(0, L, n+1)
    xp       =  xu_full[:-1] + 0.5*np.diff(xu_full)
    xu       =  xu_full[:-1]
    return xp, xu

def split_path(x):
    L =  ' / '.replace(' ','')
    W = r' \ '.replace(' ','')
    while W in x:
        x = x.replace(W,L)
    L2 = L*2
    while L2 in x:
        x = x.replace(L2,L)
    return x.split(L)

def Zipped_Folder(folder, target_fname):
    with zipfile.ZipFile(target_fname, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os_walk(folder):
            for file in files:
                file = pjoin(root, file)
                if all((not avoid in piece) for piece in split_path(file) 
                                            for avoid in ['__pycache__','.git']):
                    zipf.write(file                        , 
                               prelpath(file               , 
                                        pjoin(folder, '..')))

# def activate_omp(run_folder):
#     
#     get_pad  =  lambda x: ' '*(len(x) - len(x.lstrip(' ')))
#     comment  =  lambda x: get_pad(x) + '! ' + x.lstrip(' ')
# 
#     for fname in listdir_full_files(run_folder):
#         if any(fname.endswith(end) for end in ['.cuf','.f90','.f']):
#             A      =  reader(fname)
#             i_cuf  =  [i for i,x in enumerate(A) if '$cuf kernel'      in x if x.replace(' ','')[1] != '!']
#             i_end  =  [i for i,x in enumerate(A) if 'omp_end_parallel' in x if x.replace(' ','')[1] != '!']
#             # assert sum(('kernel' in x) for x in A) == len(i_cuf)
#             assert len(i_cuf) == len(i_end)
# 
#             for i in i_end:
#                 assert '! omp_end_parallel' in A[i]
#                 A[i] = A[i].replace('! omp_end_parallel', '!$OMP END PARALLEL DO')
#             
#             for i in i_cuf[::-1]:
#                 assert 'omp_supplement' in A[i-1]
#                 A[i-1] = comment(A[i-1])
#                 A[i  ] = comment(A[i  ])
#                 pad    = get_pad(A[i])
#                 _,new  =  A[i-1].split('omp_supplement')
#                 new    = lfilter(None, [x.lstrip(' ').rstrip(' ') for x in  new.lstrip(' :').replace(')',')\n').split('\n')])
#                 new    = [ "!$OMP   PARALLEL DO"] + \
#                          [f"!$OMP {x}" for x in new]
#                 for j in range(len(new)-1):
#                     new[j] = new[j] + ' &'
#                 A = A[:i+1] + [(pad+x) for x in new] + A[i+1:]
# 
#             # remove device declaration
#             for i in range(len(A)):
#                 if ',device::' in A[i].replace(' ',''):
#                     a,b   =  A[i].split('device')
#                     a     =  a.rstrip(' ,')
#                     A[i]  =  a+b
# 
#             writer(fname, A)

get_pad   =  lambda x: ' '*(len(x) - len(x.lstrip(' ')))

def check_split(x, tag, f):
    assert x[:len(tag)] == tag
    x0    =  x + ''
    b, x  =  x.split(tag)
    assert (not b) and (x==x0[len(tag):])
    return f(x)

class Parse_Alloc:
    def __init__(self, run_folder):
        main_dns  =  pjoin(run_folder, 'main_dns.cuf')

        assert isfile(main_dns)

        A                       =  reader(main_dns)
        all_alloc, all_dealloc  =  [], []

        for i in range(len(A)):
            if ('device::' in A[i].replace(' ','')) and (A[i].lstrip(' ')[0]!='!'):
                declare, alloc, dealloc =  self.get_device_vars(A[i])
                A[i]                    =  declare
                all_alloc  .append(alloc  )
                all_dealloc.append(dealloc)

        for tag,arr in [['PYTHON_HOLDER_ALLOCATOR'  , all_alloc  ],
                        ['PYTHON_HOLDER_DEALLOCATOR', all_dealloc]]:

            i   = locate_ind(tag, A)+1
            pad = get_pad(A[i-1])
            A = A[:i] + [pad+x for x in arr] + A[i:]

        writer(main_dns, A)
    
    @staticmethod
    def get_device_vars(x):

        clean  =  lambda x: x.split('(')[0]

        def dotted(x):
            if '(' in x:
                n = x.count(',')
                x = x.split('(')[0] + '(' + ','.join(':'*(n+1)) + ')'
            return x

        assert not ';' in x

        prefix  =  x.split('::')[0] + ',allocatable::'
        x       =  list(x.replace(' ','').split('!')[0].split('::')[1])

        lvl     =  0

        for i,c in enumerate(x):
            if c == '(':  lvl += 1
            if c == ')':  lvl -= 1
            if (lvl == 0) and (c == ','):
                x[i] = ';'
        x = (''.join(x)).split(';')

        alloc   = f"allocate({','.join(x)})"
        dealloc = f"deallocate({','.join(map(clean ,x))})"
        declare = f"{prefix}{   ','.join(map(dotted,x))}"

        return declare, alloc, dealloc

def Activate_Cuda_Id(run_folder, special_cuda_id, tag = 'PYTHON_HOLDER_setdevice'):
    main_dns  =  pjoin(run_folder, 'main_dns.cuf')
    A         =  reader(main_dns)
    i,        =  [i for i,x in enumerate(A) if tag in x]
    pad       =  get_pad(A[i])
    if special_cuda_id >=0:
        A[i]      =  pad + f'ierr  =  cudaSetDevice({special_cuda_id}) ! {tag}'
    else:
        new = [pad+"open(101,file='device_id.txt')",
               pad+"   read(101,*) device_id"      ,
               pad+"close(101)"                    ,
               pad+"ierr  =  cudaSetDevice(device_id)"]
        A = A[:i] + new + A[i:]

    writer(main_dns, A)

def apply_variable_replacements(run_folder, D):
    
    def join_continues(A):
        for i in reversed(range(len(A))):
            if A[i].split('!')[0].rstrip(' ')[-1:] == '&':
                A[i] = A[i].split('!')[0] # remove comment after &
                A[i] += '\n' + A[i+1]
                A.pop(i+1)
        return A

    has_codeline = lambda x: x.split('!')[0].lstrip(' ')

    def comment(x):
        all_x  =  x.split('\n')
        x      = [f"{get_pad(x)}! {x.lstrip(' ')}" for x in all_x]
        return '\n'.join(x)
    
    is_lhs = lambda x, name: x.split('=')[0].lstrip(' ').rstrip(' ') == name
    
    def fetch_rhs(A, name):
        found = 0 
        for k in range(len(A)):
            if is_lhs(A[k], name):
                found  +=  1
                result  =  '='.join(A[k].split('!')[0].split('=')[1:])
        assert found == 1
        return result


    for fname in lfilter(lambda x: basename(x).endswith('.cuf'), listdir_full_files(run_folder)):
        A     = reader(fname)
        all_i = [i for i,x in enumerate(A) if '#BEGIN_PYTHON_REPLACEMENT'  in x]
        all_j = [i for i,x in enumerate(A) if '#FINISH_PYTHON_REPLACEMENT' in x]
        assert len(all_i)==len(all_j)
        if all_i:
            for i,j in reversed(list(zip(all_i,all_j))):
                assert A[i].count('[') == A[i].count(']') == 1
                tags = [val.lstrip(' ').rstrip(' ') for val in A[i].split('[')[1].split(']')[0].split(';')]
                assert len(tags) in [3,4]
                if D[tags[0]]:
                    snippet      =  join_continues(A[i+1:j])
                    snippet0     =  deepcopy(snippet)
                    name, value  =  tags[1],tags[2]

                    if value == 'PYTHON_fetch_RHS':
                        value = fetch_rhs(snippet, name)
                    
                    for k in range(len(snippet)):
                        if has_codeline(snippet[k]):
                            snippet[k] = snippet[k].split('!')[0]
                            if name in snippet[k]:
                                if is_lhs(snippet[k], name):
                                    if len(tags) == 4:
                                        assert tags[3] == 'comment_var'
                                        snippet[k] = comment(snippet[k])
                                    else:
                                        assert tags[2] != 'PYTHON_fetch_RHS'
                                        snippet[k] = snippet[k].replace(name, value)
                                else:
                                    snippet[k] = snippet[k].replace(name, value)
                    if snippet != snippet0:
                        A = A[:i+1] + snippet + A[j:]
                        writer(fname,A)

# ------------------------------------------------------------------------
#                 Scan Data
# ------------------------------------------------------------------------

class Scan_Data:
    def __init__(self, folder):
        
        self.scan_params(folder)

        self.fname_array_ijkv            =  locate_one('array_ijkv_'             ,
                                                       listdir_full_files(folder))
        
        self.fname_array_ijkuwt          =  locate_one('array_ijkuwt_'           ,
                                                       listdir_full_files(folder))
        
        self.fname_array_ijkRuwt         =  locate_one('array_ijkRuwt_'          ,
                                                       listdir_full_files(folder))  if self.read_R_olds else ''
        
        self.fname_array_ijkRv           =  locate_one('array_ijkRv_'             ,
                                                       listdir_full_files(folder))  if self.read_R_olds else ''
        
        self.fname_array_ijk_rhs_custom  =  locate_one('array_ijk_rhs_custom_'   ,
                                                       listdir_full_files(folder))  if self.rhs_custom  else ''

        self.import_y(folder)

        fname_surf = lfilter(lambda x: 'rough_surf_define.py' == basename(x) ,
                             listdir_full_files(folder)                      )

        if self.use_rough_surface:
            self.import_surf(pop1(fname_surf))
        else:
            pass
            # assert not fname_surf
    
    def import_surf(self, fname_surf):
        
        sys_path.append(dirname(fname_surf))
        from rough_surf_define import Height_Function
        sys_path.pop()

        surf        =  Height_Function()
        self.get_h  =  surf.get_h

        assert abs(self.Lx - surf.Lx) < 1e-10
        assert abs(self.Lz - surf.Lz) < 1e-10
        assert abs(self.Ly - surf.Ly) < 1e-10
        
        self.xp, self.xu  =  get_coords_pu(self.Lx, self.nx)
        self.zp, self.zu  =  get_coords_pu(self.Lz, self.nz)

    def import_y(self, folder):
        self.yu   =  np.array(lmap(float,lfilter(None, reader(locate_one('yu.txt'        ,
                                                               listdir_full_files(folder)))))).reshape(-1)
        assert len(self.yu) == (self.ny+1)
        # self.yu          -=  self.yu.min()
        self.yp           =  self.yu[:-1] + 0.5*np.diff(self.yu)
        self.yp_extended  =  [float(self.yu[0])] + self.yp.tolist() + [float(self.yu[-1])]
        self.yu_extended  =  ([float(2*self.yu[ 0]-self.yu[ 1])]    + self.yu.tolist() + 
                              [float(2*self.yu[-1]-self.yu[-2])]                       )

        assert len(self.yp) == self.ny
    
    def scan_params(self, folder):
        
        D_param                     =  import_pyjson(locate_one('param.txt'                ,
                                                                listdir_full_files(folder)))
    
        # ---------------- domain parameters ----------------

        self.nx          =  int(  D_param['nx'])
        self.nz          =  int(  D_param['nz'])
        self.ny          =  int(  D_param['ny'])

        self.Lx          =  float(D_param['Lx'])
        self.Lz          =  float(D_param['Lz'])
        self.Ly          =  float(D_param['Ly'])

        self._Ret        =  float(D_param['Ret'])
        self._Pr         =  float(D_param['Pr'])

        self.Sf_x        =  float(D_param['Sf_x'])
        self.Sf_z        =  float(D_param['Sf_z'])

        assert abs(self.Sf_x**2 + self.Sf_z**2 - 1) < 1e-10

        self.dtmax       =  float(D_param['dtmax'])

        self.nu          =  1./self._Ret
        self.cond        =  1./self._Ret/self._Pr

        # ---------------- general options ----------------
        self.use_rough_surface  =  D_param['use_rough_surface']
        
        self.nprint             =  int (D_param['nprint'])

        self.solve_thermal      =  D_param['solve_thermal']
        self.write_end_file     =  D_param['write_end_file']
        self.read_R_olds        =  D_param['read_R_olds']

        # ---------------- averaging options ----------------
        self.track_budgets    =  D_param['track_budgets']
        self.avg_on_gpu       =  D_param['avg_on_gpu']

        self.start_avg        =  int (D_param['start_avg'])
        self.num_write_avg    =  int (D_param['num_write_avg'])

        self.freq_avg         =  int (D_param['freq_avg'])
        self.N_raw_avg        =  int (D_param['N_raw_avg'])

        self.tree_avg         =       D_param['tree_avg']
        self.nbins_tree       =  int (D_param['nbins_tree'])

        if self.tree_avg:
            assert self.nbins_tree >= 1
        else:
            assert self.nbins_tree <= 0
        
        # ---------------- filtering (optional) ----------------
        
        self.apply_filtering       =      D_param['apply_filtering']
        
        if self.apply_filtering:
            self.steps_filtering   =  int(D_param['steps_filtering'])
            self.n_stop_filtering  =  int(D_param['n_stop_filtering'])
            self.first_filtering   =  int(D_param['first_filtering'])

        # ---------------- validation options (do not use) ----------------
        self.rhs_custom         =  D_param['rhs_custom']
        self.verif_output       =  D_param['verif_output']

        self.use_rough_ghost    =  D_param['use_rough_ghost_cells']

        # if self.use_rough_ghost:
        #     self.ghost_adim_tol =  float(D_param['ghost_adim_tol'])

        assert (self.use_rough_ghost + self.use_rough_surface) in [0,2] # both True or both False

# ------------------------------------------------------------------------
#                  Global Configuration
# ------------------------------------------------------------------------
# global_stack = []
def dns_configure(base_folder, special_cuda_id, manual_include, no_comment):

    info    =  Scan_Data(pjoin(base_folder,'input'))

    assert basename(base_folder).endswith('_vrun')
    
    # as_fortran_logical  =  lambda x: '.true.' if x else '.false.'

    holders = dict(PYTHON_HOLDER_Lx                    =  info.Lx                                                        ,
                   PYTHON_HOLDER_Lz                    =  info.Lz                                                        ,
                   PYTHON_HOLDER_nu                    =  info.nu                                                        ,
                   PYTHON_HOLDER_cond                  =  info.cond                                                      ,
                   PYTHON_HOLDER_dtmax                 =  info.dtmax                                                     ,
                   PYTHON_HOLDER_nx                    =  info.nx                                                        ,
                   PYTHON_HOLDER_ny                    =  info.ny                                                        ,
                   PYTHON_HOLDER_Sf_x                  =  info.Sf_x                                                      ,
                   PYTHON_HOLDER_Sf_z                  =  info.Sf_z                                                      ,
                   PYTHON_HOLDER_nz                    =  info.nz                                                        ,
                   PYTHON_HOLDER_array_ijkv            =  basename(info.fname_array_ijkv  ).removesuffix('.zip')         ,
                   PYTHON_HOLDER_array_ijkuwt          =  basename(info.fname_array_ijkuwt).removesuffix('.zip')         ,
                   PYTHON_HOLDER_nbins_tree            =  info.nbins_tree                                                ,
                   PYTHON_HOLDER_N_raw_avg             =  info.N_raw_avg                                                 ,
                   PYTHON_HOLDER_start_avg             =  info.start_avg                                                 ,
                   PYTHON_HOLDER_num_write_avg         =  info.num_write_avg                                             ,
                   PYTHON_HOLDER_freq_avg              =  info.freq_avg                                                  ,
                   PYTHON_HOLDER_nprint                =  info.nprint                                                    ,
                   PYTHON_HOLDER_array_ijkRuwt         =  basename(info.fname_array_ijkRuwt       ).removesuffix('.zip') ,
                   PYTHON_HOLDER_array_ijkRv           =  basename(info.fname_array_ijkRv         ).removesuffix('.zip') ,
                   PYTHON_HOLDER_array_ijk_rhs_custom  =  basename(info.fname_array_ijk_rhs_custom).removesuffix('.zip') ,
                   PYTHON_HOLDER_expfolder             =  basename(base_folder)                                          ,
                   )
    if info.apply_filtering:
        holders.update(dict(PYTHON_HOLDER_steps_filtering       =  info.steps_filtering                       ,
                            PYTHON_HOLDER_n_stop_filtering      =  info.n_stop_filtering                      ,
                            PYTHON_HOLDER_first_filtering       =  info.first_filtering                       ))
    defined  =  dict(PYTHON_DEF_track_turb_budgets = info.track_budgets                  ,
                     PYTHON_DEF_verif_output       = info.verif_output                   ,
                     PYTHON_DEF_binary_tree_avg    = info.tree_avg                       ,
                     PYTHON_DEF_avg_on_gpu         = info.avg_on_gpu                     ,
                     PYTHON_DEF_solve_thermal      = info.solve_thermal                  ,
                     PYTHON_DEF_rhs_custom         = info.rhs_custom                     ,
                     PYTHON_DEF_read_R_olds        = info.read_R_olds                    ,
                     PYTHON_DEF_use_rough_surface  = info.use_rough_surface              ,
                     PYTHON_DEF_write_end_file     = info.write_end_file                 ,
                     PYTHON_DEF_rough_ghostcells   = info.use_rough_ghost                ,
                     PYTHON_DEF_apply_filtering    = info.apply_filtering                ,
                     )
    
    run_folder = pjoin(base_folder, 'runtime')

    mkdir_p(run_folder)
    mkdir_p(pjoin(run_folder, 'AVG'      ))
    mkdir_p(pjoin(run_folder, 'geom_vars'))
    mkdir_p(pjoin(run_folder, 'END'      ))
    mkdir_p(pjoin(run_folder, 'MOVIE_I_MID'))
    mkdir_p(pjoin(run_folder, 'MOVIE_J_MID'))
    mkdir_p(pjoin(run_folder, 'MOVIE_K_MID'))

    if info.verif_output:
        mkdir_p(pjoin(run_folder,'VERIF'))
    
    fill_geom_vars( info.yu          ,
                    info.yp          ,
                    info.yp_extended ,
                    info.ny          ,
                   (info.nx,info.nz) ,
                   pjoin(run_folder, 'geom_vars'),
                   )
    
    cuf_template_folder  =  pjoin(deepdirname(abspath(__file__),2), 'cuf_code')
    
    main_dns             =  pjoin(run_folder, 'main_dns.cuf')

    os_system(f"cp {pjoin(cuf_template_folder, 'template_dns.cuf')} {main_dns}")

    if info.use_rough_surface:
        make_rough_surface(info, run_folder)
        if info.use_rough_ghost:
            # global_stack.extend([info, run_folder]) ; 1/0
            holders.update(make_rough_surface_ghost(info, run_folder))

    for include_file in lfilter(lambda x: not 'template_dns.cuf' in basename(x),
                                listdir_full_files(cuf_template_folder)        ):
        
        new_file  =  pjoin(run_folder, basename(include_file))

        os_system(f"cp {include_file} {new_file}")
        writer(new_file, resolve_define(reader(new_file), defined, must_find = False))
        quick_replacer(new_file, holders, None)
    
    os_system(f"cp -R {pjoin(cuf_template_folder,'post')} {pjoin(run_folder,'post')}")

    quick_replacer(main_dns, holders, defined)
    
    # writer(pjoin(run_folder, 'geom_vars.dat'), data_geom_vars)
    
    if special_cuda_id:
        Parse_Alloc(run_folder)
        Activate_Cuda_Id(run_folder, special_cuda_id)

    if manual_include:
        apply_include(main_dns)

    # apply_variable_replacements(run_folder, dict(PYTHON_HOLDER_nonzero_deviceid = (special_cuda_id != 0)))
    if no_comment:
        apply_no_comment(main_dns)
    
    Zipped_Folder(dirname(cuf_template_folder), pjoin(base_folder, 'backup_source.zip'))

# ------------------------------------------------------------------------
#                  Direct Runner
# ------------------------------------------------------------------------

if __name__ == '__main__':
    
    assert len(argv) in [2,3]
    dns_configure(             argv[1]                                                ,
                   check_split(argv[2],'CUDA_SET_DEVICE_',int) if len(argv)==3 else 0 ,
                   manual_include = True                                             ,
                   no_comment = True)


    
