# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  sympy                    as  sp
from    pprint  import  pformat 
import  numpy                    as  np

# ------------------------------------------------------------------------
#                 Utilities
# ------------------------------------------------------------------------

def str_dict_aligned(d, all_lamstr = [], marker = '#', register_marker='&'):
    assert type(d) == dict
    assert not marker          in str(d)
    assert not register_marker in str(d)

    register  =  {}

    def dfs(d):
       if type(d) != dict:
        new_val = register_marker+str(len(register))+register_marker
        register[new_val] = d
        return new_val
       L       =  max(map(len,d.keys()))
       result  =  {}
       for key, value in d.items():
           result[key + marker*(L - len(key))] = dfs(value)
       return result
    
    d            =  list(pformat(dfs(d), width=1, sort_dicts=False, indent=2))

    find_inds    =  lambda c: [i for i,x in enumerate(d) if x==c]
    inds_marker  =  find_inds(marker)
    inds_dot     =  find_inds(':')

    for i in inds_marker[::-1]:
       j = list(filter(lambda j: i<j, inds_dot))[0]
       d.insert(j+1, ' ')
       d.pop(i)
    
    d = ''.join(d)

    for key,val in register.items():
        d = d.replace(f"'{key}'", str(val))
    
    for lamstr in all_lamstr:
        if lamstr:
            d = d.replace(lamstr + "':" , "':" + lamstr)
            for i in range(100):
                d = d.replace(lamstr + ' '*i          +  '{',
                                       ' '*i + lamstr +  '{')
        
    return d

# ------------------------------------------------------------------------
#                  Interpolation Coefficients and Derivatives
# ------------------------------------------------------------------------

def get_interpolation_coefficients(output = False):
    result = {}
    all_lamstr = {2: ' lambda xc,x0,x1,x2   :',
                  3: ' lambda xc,x0,x1,x2,x3:'}
    x0,x1,x2,x3 = sp.symbols('x0,x1,x2,x3', real=True)
    f_ind  =  lambda label, order: [f"{label}_{i}" for i in [0,order]]
    def add_order2():
        order = 2
        result[str(order)] = {}
        for label in ['normal',*f_ind('dirichlet',order),*f_ind('neumann',order)]:

            x,a,b,c,d,xc,y0,y1,y2,y3 = sp.symbols('x,a,b,c,d,xc,y0,y1,y2,y3', real=True)

            if order == 2:
                d = 0

            f   = d*(x**3) + a*(x**2) + b*x + c

            def mk_eq(pos):
                xv = [x0,x1,x2,x3][pos]
                yv = [y0,y1,y2,y3][pos]
                if   f"dirichlet_{pos}" == label: return f        .subs(x,xv)
                elif f"neumann_{pos}"   == label: return f.diff(x).subs(x,xv)
                else:                             return f        .subs(x,xv) - yv 

            eqs     =  list(map(mk_eq,range(order+1)))
            myvars  =  [a,b,c,d][:order+1]
            sol     =  sp.solve(eqs, *myvars)
            f_sol   =  f
            for myvar in myvars:
                f_sol   = f_sol.subs(myvar, sol[myvar])
            
            f       =  f_sol                 .subs(x,xc)
            dfdx    =  f_sol.diff(x)         .subs(x,xc)
            d2fdx2  =  f_sol.diff(x).diff(x) .subs(x,xc)

            lamstr  =  all_lamstr[order]

            def mk_str(f):
                new =             {'p0'+lamstr:    f.subs(y0,1).subs(y1,0).subs(y2,0).subs(y3,0),
                                   'p1'+lamstr:    f.subs(y0,0).subs(y1,1).subs(y2,0).subs(y3,0),
                                   'p2'+lamstr:    f.subs(y0,0).subs(y1,0).subs(y2,1).subs(y3,0)}
                if order == 3: new['p3'+lamstr] =  f.subs(y0,0).subs(y1,0).subs(y2,0).subs(y3,1)
                return new

            result[str(order)][label] = {  'f'   : mk_str(  f   ) ,
                                          'dfdx' : mk_str( dfdx ) ,
                                         'd2fdx2': mk_str(d2fdx2)}
    add_order2()

    def add_order3():
        result['3'] = {}
        def s(p, mode, all_x):
            if   '_0' in label:
                lbl = {'a': label   , 'b':'normal'}[mode]
            elif '_3' in label:
                lbl = {'a': 'normal', 'b':label.replace('_3','_2')}[mode]
            else:
                assert label == 'normal'
                lbl = label
            c0, c1, c2  =  all_x
            ans         = result['2'][lbl][subf][f'p{p}'+all_lamstr[2]].subs(x2,c2).subs(x1,c1).subs(x0,c0)
            return f'(({ans})/2)'
        
        for label in ['normal',*f_ind('dirichlet',3),*f_ind('neumann',3)]:
            result['3'][label] = {}
            for subf in ['f','dfdx','d2fdx2']:
                result['3'][label][subf] = {}
                result['3'][label][subf]['p0'+all_lamstr[3]] =  s(0,'a',(x0,x1,x2))
                result['3'][label][subf]['p1'+all_lamstr[3]] =  s(1,'a',(x0,x1,x2))+ ' + ' + s(0,'b',(x1,x2,x3))
                result['3'][label][subf]['p2'+all_lamstr[3]] =  s(2,'a',(x0,x1,x2))+ ' + ' + s(1,'b',(x1,x2,x3))
                result['3'][label][subf]['p3'+all_lamstr[3]] =                               s(2,'b',(x1,x2,x3))
    add_order3()

    result = str_dict_aligned(result, all_lamstr = list(all_lamstr.values()))

    if output:
        print(result)

    return result

# ------------------------------------------------------------------------
#                  Super Bands (geom_vars)
# ------------------------------------------------------------------------

def mk_geom_vars(y_distrib, type_bc, y_base = None, order = 2, coeffs = eval(get_interpolation_coefficients())):
    if y_base is None:
        y_base = y_distrib
    
    assert type_bc in ['normal', 'dirichlet', 'neumann']
    assert order   in [2,3]
    all_c      = 'abcd'[:order+1]
    geom_vars  =  {f'inds_{c}':[] for c in all_c}

    # 0->a
    # 1->b
    # 2->c
    
    for i in range(len(y_distrib)):
        
        if   order == 3:  all_i = np.array([i-2, i-1, i, i+1]) + 1
        elif order == 2:  all_i = np.array([     i-1, i, i+1])
        else           :  1/0

        while all_i.min() <  0          : all_i += 1
        while all_i.max() >= len(y_base): all_i -= 1

        label = 'normal'
        if type_bc != 'normal':
            for bad in [0, len(y_base)-1]:
                if bad in all_i:
                    assert label == 'normal'
                    label = f"{type_bc}_{list(all_i).index(bad)}"
        
        for p,c in enumerate(all_c):
            geom_vars[f'inds_{c}'].append(all_i[p])
        
        yc     =  0 # y_distrib[i]
        all_y  =  np.array([y_base   [j] for j in all_i]) - y_distrib[i]
        for ff in (['dfdx'] if order==3 else ['f', 'dfdx', 'd2fdx2']):
            for p,c in enumerate(all_c):
                name = {'f': 'interp', 'dfdx': 'ddy', 'd2fdx2': 'd2dy2'}[ff]
                key_gv = f'{name}_{c}'
                # print(label,ff,yc, *all_y)
                if not key_gv in geom_vars:
                    geom_vars[key_gv] = []
                geom_vars[key_gv].append(coeffs[str(order)][label][ff][f"p{p}"](yc, *all_y))
    return geom_vars

# # ------------------------------------------------------------------------
# #                  Tetrahedron Fit
# # ------------------------------------------------------------------------
# 
# def mk_tetra_fit():
#     
#     x,y,z,a,b,c,d,x0,y0,z0,f0,x1,y1,z1,f1,x2,y2,z2,f2,x3,y3,z3,f3,xx,yy,zz = sp.symbols('x,y,z,a,b,c,d,x0,y0,z0,f0,x1,y1,z1,f1,x2,y2,z2,f2,x3,y3,z3,f3,xx,yy,zz', real=True)
#     myvars                          =  a,b,c,d
#     f                               =  a*x + b*y + c*z + d
#     eqs                             =  [ f.subs(x, x0).subs(y, y0).subs(z, z0) - f0 ,
#                                          f.subs(x, x1).subs(y, y1).subs(z, z1) - f1 ,
#                                          f.subs(x, x2).subs(y, y2).subs(z, z2) - f2 ,
#                                          f.subs(x, x3).subs(y, y3).subs(z, z3) - f3 ]
#     sol                             =  sp.solve(eqs, *myvars)
#     f                               =  f.subs(a, sol[a]).subs(b, sol[b]).subs(c, sol[c]).subs(d, sol[d]).subs(x,xx).subs(y,yy).subs(z,zz).expand()
#     
#     get_s                           =  lambda val: (f - f.subs(val,0)).subs(val,1) 
#     s_f0, s_f1, s_f2, s_f3          =  lmap(get_s, [f0, f1, f2, f3])
#     
#     assert (f0*s_f0 +
#             f1*s_f1 +
#             f2*s_f2 +
#             f3*s_f3 - f).simplify() == 0
#     
#     as_f    =  lambda s: f"lambda xx,yy,zz,x0,y0,z0,x1,y1,z1,x2,y2,z2,x3,y3,z3: {s}"
#     result  =  str(dict(key_f0_key = as_f(s_f0) ,
#                         key_f1_key = as_f(s_f1) ,
#                         key_f2_key = as_f(s_f2) ,
#                         key_f3_key = as_f(s_f3) ))
#     result  =  result.replace('"','').replace("'","").replace('key_','"').replace('_key','"')
#     return result